;; 1. (1p)
;; construiti o lista din variabilele a, b, c si d folosind
;; a) exclusiv functia list
;; b) exclusiv functia cons
;; c) inclusiv functia append
(define make-list-a
  (lambda (a b c d)
    (list a b c d)))

(display "1a.  (make-list-a 1 'a 'b 'c)\n\t")
(make-list-a 1 'a 'b 'c)

;; OBSERVATIE
;; intrucat Scheme permite sintaxa simplificata (define (f x y) ...)
;; in loc de (define f (lambda (x y) ...)), de acum inainte vom 
;; defini functii folosind aceasta sintaxa

(define (make-list-b a b c d)
  (cons a (cons b (cons c (cons d '())))))

(display "1b.  (make-list-b 1 'b 'c 'a)\n\t")
(make-list-a 1 'b 'c 'a)

(define (make-list-c a b c d)
  (append (list a b) (list c d)))

(display "1c.  (make-list-c 1 'c 'a 'b)\n\t")
(make-list-a 1 'c 'a 'b)

;; OBSERVATIE
;; - nu se putea rezolva punctul c exclusiv cu append
;; - append este un operator care ia ca argumente numai liste,
;;   ori a b c d sunt numai elemente


;; 2. fibonacci recursiv (1p)
;; veti avea ocazia sa optimizati aceasta varianta ineficienta de fibonacci
(define (fibonacci n)
  (if (< n 2)
      n
      (+ (fibonacci (- n 1)) (fibonacci (- n 2)))))

(display "2.  (map fibonacci '(0 1 2 3 4 5 6 7 8 9 10))\n\t")
(map fibonacci '(0 1 2 3 4 5 6 7 8 9 10)) 
; map e o functie ce primeste ca parametri o functie si o lista si aplica functia respectiva asupra fiecarui element
; din lista

;; 3. take - primeste o lista L si un numar n si intoarce lista formata din primele n elemente din L (1p)
(define (take L n)
  (if (or (null? L) (= n 0))
      null
      (cons (car L) (take (cdr L) (- n 1)))))

(display "3. (take '(1 2 3 4 5) 2))\n\t")
(take '(1 2 3 4 5) 2)

;; 4. drop - primeste o lista L si un numar n si intoarce lista formata din elementele din L mai putin primele n (1p)
(define (drop L n)
  (if (or (null? L) (= n 0))
      L
      (drop (cdr L) (- n 1))))

(display "4. (drop '(1 2 3 4 5) 2))\n\t")
(drop '(1 2 3 4 5) 2)

;; 5. my-reverse - inverseaza o lista (se va implementa obligatoriu recursiv) (1p)
(define (my-reverse L)
  (if (null? L)
      L
      (append (my-reverse (cdr L))
              (list (car L)))))

(display "5.  (my-reverse '(s t r e s s e d))\n\t")
(my-reverse '(s t r e s s e d))
(display "     at least it works for me :-)\n")

;; OBSERVATIE
;; - aici se justifica (list (car L)) sau varianta mai putin estetica
;;   dar echivalenta (cons (car L) '()), deoarece lista de un element
;;   rezultata se pune LA COADA unei alte liste  
;; - pentru a pune un element in fata unei liste L s-ar fi folosit 
;;   (cons elem L) (mult mai estetic decat (append (list elem) L))


;; 6. deep-reverse - inverseaza o lista, actionand inclusiv asupra elementelor din eventualele liste interioare 
;; (se va implementa obligatoriu recursiv) (2p)
(define (deep-reverse L)
  (cond
    ((null? L) L)
    ((list? (car L)) (append (deep-reverse (cdr L))
                             (list (deep-reverse (car L)))))
    (else (append (deep-reverse (cdr L))
                  (list (car L))))))

(display "6.  (deep-reverse '(1 (2 (3 4) 5 (6 ((7) 8) 9)) 10))\n\t")
(deep-reverse '(1 (2 (3 4) 5 (6 ((7) 8) 9)) 10))

;; OBSERVATIE
;; este aceeasi prelucrare cu cea de la my-reverse, cu o singura 
;; exceptie: cand (car L) este o lista, acesteia i se aplica recursiv 
;; deep-reverse inainte de a fi adaugata la rezultat


;; 7. merge - interclaseaza 2 liste sortate crescator (2p)
(define (merge L1 L2)
  (cond ((null? L1) L2)
        ((null? L2) L1)
        ((< (car L1) (car L2)) (cons (car L1) (merge (cdr L1) L2)))
        (else (cons (car L2) (merge L1 (cdr L2))))))

(display "7. (merge '(1 4 6 8) '(2 3 5 6 7))\n\t")
(merge '(1 4 6 8) '(2 3 5 6 7))


;; 8. merge-sort - sorteaza o lista crescator folosind algoritmul merge-sort (3p)
;; (se va implementa obligatoriu recursiv si NU se va utiliza functia deja existenta mergesort :P)
;; Hint: Folositi take si drop pentru a separa o lista in prima jumatate si a doua jumatate
(define (merge-sort L)
  (if (or (null? L) (null? (cdr L)))
      L
      (merge (merge-sort (take L (quotient (length L) 2))) (merge-sort (drop L (quotient (length L) 2))))))

(display "8. (merge-sort '(6 5 4 3 2 1))\n\t")
(merge-sort '(6 5 4 3 2 1))

;; un mic benchmark
;; puteti folosi aceasta functie pentru a testa performantele
;; diversilor algoritmi de sortare pe liste de dimensiuni mari
;; ex: (time (merge-sort (make-random-list 10000)))
(define make-random-list
  (lambda (n)
    (if (zero? n)
        '()
        (cons (random 100) (make-random-list (- n 1))))))

(display "8.  (merge-sort (make-random-list 25))\n\t")
(merge-sort (make-random-list 25))